Namespace Anova
  Friend Class AnovaProgram
	Shared Sub Main(ByVal args() As String)
	  Console.WriteLine(vbCrLf & "Begin ANOVA using VB demo " & vbCrLf)
	  Console.WriteLine("The goal is to infer if 3 or more population")
	  Console.WriteLine("means are all equal based on samples. " & vbCrLf)

	  Console.WriteLine("The sample data is: " & vbCrLf)

            Dim data(2)() As Double  ' 3 groups
            data(0) = New Double() { 3, 4, 6, 5 }
	  data(1) = New Double() { 8, 12, 9, 11, 10, 8 }
	  data(2) = New Double() { 13, 9, 11, 8, 12 }
	  Dim colNames() As String = { "Group1", "Group2", "Group3" }
	  ShowData(data, colNames)

	  Console.WriteLine(vbCrLf & "Calculating F-statistic")
	  Dim df() As Integer = Nothing ' degrees of freedom
	  Dim F As Double = Fstat(data, df)
	  Console.WriteLine("F = " & F.ToString("F3"))

	  Console.Write("The degrees of freedom are ")
	  Console.WriteLine(df(0) & ", " & df(1))

	  Console.WriteLine(vbCrLf & "Calculating p-value")
	  Dim pValue As Double = QF(df(0), df(1), F)
	  Console.Write("p-value = ")
	  Console.WriteLine(pValue.ToString("F8"))

	  Console.WriteLine(vbCrLf & "The p-value is the probability that all ")
	  Console.WriteLine("group means are equal.")
	  Console.WriteLine("Interpreting the p-value is subtle.")

	  Console.WriteLine(vbCrLf & "End ANOVA demo " & vbCrLf)
	  Console.ReadLine()
	End Sub ' Main

	Private Shared Function Fstat(ByVal data()() As Double, <System.Runtime.InteropServices.Out> ByRef df() As Integer) As Double
	  ' calculate F statistic and degrees freedom num, denom
	  ' assumes data has specific structure:
	  ' data0 -> 3, 4, etc (group 1)
	  ' data1 -> 8, 12, etc. (group 2)
	  ' etc.

	  Dim K As Integer = data.Length ' number groups
	  Dim n(K - 1) As Integer ' number items each group
	  Dim N_Renamed As Integer = 0 ' total number data points
	  For i  = 0 To K - 1
		n(i) = data(i).Length
		N_Renamed += data(i).Length
	  Next i

	  ' 1. group means and grand mean
	  Dim means(K - 1) As Double
	  Dim gMean As Double = 0.0
	  For i  = 0 To K - 1
		For j  = 0 To data(i).Length - 1
		  means(i) += data(i)(j)
		  gMean += data(i)(j)
		Next j
		means(i) /= n(i)
		Console.WriteLine("Group [" & i & "] mean = " & means(i).ToString("F2"))
	  Next i
	  gMean /= N_Renamed
	  Console.WriteLine("Overall mean = " & gMean.ToString("F2"))

	  ' 2. SSb and MSb
	  Dim SSb As Double = 0.0
	  For i  = 0 To K - 1
		SSb += n(i) * (means(i) - gMean) * (means(i) - gMean)
	  Next i
	  Dim MSb As Double = SSb / (K - 1)

	  ' 3. SSw and MSw
	  Dim SSw As Double = 0.0
	  For i  = 0 To K - 1
		For j  = 0 To data(i).Length - 1
		  SSw += (data(i)(j) - means(i)) * (data(i)(j) - means(i))
		Next j
	  Next i
	  Dim MSw As Double = SSw / (N_Renamed - K)

	  ' for demo only
	  Console.WriteLine("Calculated SSb = " & SSb.ToString("F4"))
	  Console.WriteLine("Calculated MSb = " & MSb.ToString("F4"))
	  Console.WriteLine("Calculated SSw = " & SSw.ToString("F4"))
	  Console.WriteLine("Calculated MSw = " & MSw.ToString("F4"))
	  Console.Write("F = MSb / MSw")
	  Console.WriteLine(" = " & MSb.ToString("F4") & " / " & MSw.ToString("F4"))
	  'Console.WriteLine("");

	  df = New Integer(1){} ' store df values
	  df(0) = K - 1
	  df(1) = N_Renamed - K

	  Dim F As Double = MSb / MSw
	  Return F
	End Function ' Fstat

	Private Shared Function LogGamma(ByVal z As Double) As Double
	  ' Lanczos approximation g=5, n=7
	  Dim coef() As Double = { 1.000000000190015, 76.18009172947146, -86.50532032941677, 24.01409824083091, -1.231739572450155, 0.1208650973866179e-2, -0.5395239384953e-5 }

	  Dim LogSqrtTwoPi As Double = 0.91893853320467274178

	  If z < 0.5 Then ' Gamma(z) = Pi / (Sin(Pi*z))* Gamma(1-z))
		Return Math.Log(Math.PI / Math.Sin(Math.PI * z)) - LogGamma(1.0 - z)
	  End If

	  Dim zz As Double = z - 1.0
	  Dim b As Double = zz + 5.5 ' g + 0.5
	  Dim sum As Double = coef(0)

	  For i  = 1 To coef.Length - 1
		sum += coef(i) / (zz + i)
	  Next i

	  Return (LogSqrtTwoPi + Math.Log(sum) - b) + (Math.Log(b) * (zz + 0.5))
	End Function

	Private Shared Function BetaInc(ByVal a As Double, ByVal b As Double, ByVal x As Double) As Double
	  ' Incomplete Beta function
	  ' A & S 6.6.2 and 26.5.8
	  Dim bt As Double
	  If x = 0.0 OrElse x = 1.0 Then
		bt = 0.0
	  Else
		bt = Math.Exp(LogGamma(a + b) - LogGamma(a) - LogGamma(b) + a * Math.Log(x) + b * Math.Log(1.0 - x))
	  End If

	  If x < (a + 1.0) / (a + b + 2.0) Then
		Return bt * BetaIncCf(a, b, x) / a
	  Else
		Return 1.0 - bt * BetaIncCf(b, a, 1.0 - x) / b
	  End If
	End Function

	Private Shared Function BetaIncCf(ByVal a As Double, ByVal b As Double, ByVal x As Double) As Double
	  ' Approximate Incomplete Beta computed by
	  ' continued fraction
	  ' A & S 26.5.8 
	  Dim max_it As Integer = 100
	  Dim epsilon As Double = 3.0e-7
	  Dim small As Double = 1.0e-30

	  Dim m2 As Integer ' 2*m
	  Dim aa, del As Double

	  Dim qab As Double = a + b
	  Dim qap As Double = a + 1.0
	  Dim qam As Double = a - 1.0
	  Dim c As Double = 1.0
	  Dim d As Double = 1.0 - qab * x / qap
	  If Math.Abs(d) < small Then
		  d = small
	  End If
	  d = 1.0 / d
	  Dim result As Double = d

	  Dim m As Integer
	  For m = 1 To max_it
		m2 = 2 * m
		aa = m * (b - m) * x / ((qam + m2) * (a + m2))
		d = 1.0 + aa * d
		If Math.Abs(d) < small Then
			d = small
		End If
		c = 1.0 + aa / c
		If Math.Abs(c) < small Then
			c = small
		End If
		d = 1.0 / d
		result *= d * c
		aa = -(a + m) * (qab + m) * x / ((a + m2) * (qap + m2))
		d = 1.0 + aa * d
		If Math.Abs(d) < small Then
			d = small
		End If
		c = 1.0 + aa / c
		If Math.Abs(c) < small Then
			c = small
		End If
		d = 1.0 / d
		del = d * c
		result *= del
		If Math.Abs(del - 1.0) < epsilon Then
			Exit For
		End If
	  Next m ' for

	  If m > max_it Then
		Throw New Exception("BetaIncCf failure ")
	  End If
	  Return result
	End Function ' BetaCf

	Private Shared Function PF(ByVal a As Double, ByVal b As Double, ByVal x As Double) As Double
	  ' approximate lower tail of F-dist
	  ' (area from 0.0 to x)
	  ' equivalent to the R pf() function
	  ' only accurate to about 8 decimals
	  Dim z As Double = (a * x) / (a * x + b)
	  Return BetaInc(a / 2, b / 2, z)
	End Function

	Private Shared Function QF(ByVal a As Double, ByVal b As Double, ByVal x As Double) As Double
	  ' area from x to +infinity under F
	  ' for ANOVA = prob(all means equal)
	  Return 1.0 - PF(a, b, x)
	End Function

        Private Shared Sub ShowData(ByVal data()() As Double, ByVal colNames() As String)
            For i = 0 To data.Length - 1
                Console.Write(colNames(i) & ": ")
                For j = 0 To data(i).Length - 1
                    Console.Write(data(i)(j).ToString("F2").PadLeft(7))
                Next j
                Console.WriteLine("")
            Next i
        End Sub
    End Class ' Program
End Namespace ' ns
